// TODO: put these and other similar stuff in some kind of data file

// some default stats for various classes
// add_range is the range from edges of circles, rather than centres
default_stats = {
  soldier : {
    hp : 140,
    start_mp : 0,
    atk : 10,
    asp : 0.6,
    def : 13,
    spd : 1.0,
    add_range : 0.01,
    atk_type : 'melee',
  },
  berserker : {
    hp : 100,
    start_mp : 0,
    atk : 20,
    asp : 0.8,
    def : 10,
    spd : 1.0,
    add_range : 2.0,
    atk_type : 'melee',
    proj_spd : 4.0,
  },
  archer : {
    hp : 80,
    start_mp : 0,
    atk : 15,
    asp : 1.0,
    def : 8,
    spd : 1.0,
    add_range : 2.0,
    atk_type : 'ranged',
    proj_spd : 4.0,
  },
  bmage : {
    hp : 70,
    start_mp : 0,
    atk : 5,
    asp : 0.5,
    def : 7,
    spd : 1.0,
    add_range : 2.0,
    atk_type : 'ranged',
    proj_spd : 1.0,
  },
};

class Unit {
  constructor (name, level, clss, attr, spell) {
    this.name = name;
    this.level = level;
    this.clss = clss;
    this.attr = attr;
    this.spell = spell;
    this.stats = Object.assign({}, default_stats[clss]);
  }
}

// active unit, adding battle context to the unit
class AUnit {
  constructor (unit, gx, gy, isplayer) {
    this.unit = unit;
    this.isplayer = isplayer;

    // battle-specific mutable state
    this.hp = unit.stats.hp;
    this.mp = unit.stats.start_mp;
    this.last_aa = -Infinity;
    this.isalive = true;

    this.radius = 0.5; // radius as a factor of gsize
    this.pos = [gx + 0.5, gy + 0.5];

    // TODO: factor in attribute to this
    this.bgcolor = isplayer ? "#D0D0D0" : "#000000";
    this.fgcolor = isplayer ? "#000000" : "#D0D0D0";
    this.healthcolor = isplayer ? "#00A000" : "#A00000";
  }

  pcoords () {
    return [st.grid_loc[0] + this.x() * st.gsize,
            st.grid_loc[1] + this.y() * st.gsize];
  }

  draw () {
    ctx.save();
    let pc = this.pcoords();
    ctx.translate(pc[0], pc[1]);

    // draw the circle part
    ctx.fillStyle = this.bgcolor;
    ctx.strokeStyle = this.fgcolor;
    ctx.beginPath();
    ctx.arc(0, 0, this.radius*st.gsize, 0, 2*Math.PI, false);
    ctx.stroke();
    ctx.fill();
    if ('image' in this.unit) {
      console.log('image drawing not implemented yet');
    } else {
      ctx.fillStyle = this.fgcolor;
      ctx.font = '20px Courier New';
      ctx.fillText(this.unit.clss[0] + this.unit.level, -12, 5);
    }
    ctx.stroke();
    ctx.restore();
  }

  draw_health () {
    ctx.save();
    let pc = this.pcoords();
    ctx.translate(pc[0], pc[1]);
    ctx.strokeStyle = '#000000';
    ctx.lineWidth = 2;
    ctx.fillStyle = this.healthcolor;
    ctx.globalAlpha = 1.0;
    let hbwidth = 0.8 * st.gsize;
    let hbheight = 0.15 * st.gsize;
    let tlx = -hbwidth/2;
    let tly = -this.radius * st.gsize - hbheight - 0.1*st.gsize;
    ctx.beginPath();
    ctx.rect(tlx, tly, hbwidth, hbheight);
    ctx.fillRect(tlx, tly, (this.hp / this.unit.stats.hp) * hbwidth, hbheight);
    ctx.stroke();
    ctx.restore();
  }

  draw_target () {
    if (this._target !== undefined) {
      // config
      let tail_semiangle = Math.PI / 4;
      let tail_length = 0.1 * st.gsize;

      ctx.save();
      let pc = this.pcoords();
      ctx.translate(pc[0], pc[1]);
      let tpc = vsub(this._target.pcoords(), pc);
      let left_tail = vadd(tpc, vscale(tail_length, rot(vnormalize(tpc), Math.PI - tail_semiangle)));
      let right_tail = vadd(tpc, vscale(tail_length, rot(vnormalize(tpc), Math.PI + tail_semiangle)));
      ctx.beginPath();
      ctx.strokeStyle = this.isplayer ? '#00D000' : '#D00000';
      ctx.lineWidth = 2;
      ctx.moveTo(0, 0);
      ctx.lineTo(tpc[0], tpc[1]);
      ctx.lineTo(left_tail[0], left_tail[1]);
      ctx.moveTo(tpc[0], tpc[1]);
      ctx.lineTo(right_tail[0], right_tail[1]);
      ctx.stroke();

      ctx.restore();
    }
  }

  can_hit (au) {
    if(!au) return false;
    return this.unit.stats.add_range + this.radius + au.radius >= udist(this, au);
  }

  x () {
    return this.pos[0];
  }

  y () {
    return this.pos[1];
  }

  die () {
    this.isalive = false;
    st.battle.aunits.splice(st.battle.aunits.indexOf(this), 1);
    console.log('Unit "' + this.unit.name + '" died.');
  }

  take_dmg (dmg) {
    this.hp = this.hp - dmg;
    if (this.hp < 0 && this.isalive) {
      this.die();
    }
  }


  aa (target) {
    console.log('Unit "' + this.unit.name + ' attacks "' + target.unit.name + '"');
    if (this.unit.stats.atk_type === 'melee') {
      let dmg = aa_dmg(this, target);
      target.take_dmg(dmg);

      // make the attack animation effect
      let effect = {
        type : 'melee_aa',
        user : this,
        target : target,
        start : st.frame,
        timeout : st.frame + this.unit.stats.asp * st.fps / 2,
      };
      effect.draw = function (me) {
        ctx.save();
        let pc = me.user.pcoords();
        ctx.translate(pc[0], pc[1]);
        let dir = vnormalize(vsub(me.target.pos, me.user.pos));
        let edge = vscale(me.user.radius*st.gsize, dir);
        let progress = (st.frame - me.start) / (me.timeout - me.start);
        let ext = Math.min(1, progress * 2);
        let line_dest = vscale(1+ext, edge);
        ctx.beginPath();
        ctx.strokeStyle = me.user.bgcolor;
        ctx.lineWidth = 5;
        ctx.moveTo(edge[0], edge[1]);
        ctx.lineTo(line_dest[0], line_dest[1]);
        ctx.stroke();
        ctx.restore();
      }
      st.effects.push(effect);
    } else if (this.unit.stats.atk_type === 'ranged') {
      let effect = {
        type : 'melee_aa',
        user : this,
        pos : this.pos,
        target : target,
        proj_len : 0.6,
        start : st.frame,
        timeout : Infinity,
      };

      effect.tick = function (me) {
        if (!me.target.isalive) {
          me.timeout = 0;
        }
        if (me.timeout === 0) return;

        let mvmt = me.user.unit.stats.proj_spd / st.fps;
        if (dist(me.pos, me.target.pos) < mvmt) {
          me.pos = me.target
        } else {
          me.pos = vadd(me.pos, vscale(mvmt, vnormalize(vsub(me.target.pos, me.pos))));
        }

        if (dist(me.pos, me.target.pos) < me.target.radius) {
          me.timeout = 0;
          let dmg = aa_dmg(me.user, me.target);
          target.take_dmg(dmg);
        }
      };

      effect.draw = function (me) {
        if (dist(me.target.pos, me.pos) === 0) return;
        ctx.save();
        let dir = vnormalize(vsub(me.target.pos, me.pos));
        let proj_start = [st.grid_loc[0] + me.pos[0]*st.gsize,
                          st.grid_loc[1] + me.pos[1]*st.gsize];
        let proj_end = vadd(proj_start, vscale(me.proj_len*st.gsize, dir));
        ctx.beginPath();
        ctx.moveTo(proj_start[0], proj_start[1]);
        ctx.lineTo(proj_end[0], proj_end[1]);
        ctx.stroke();
        ctx.restore();
      };
      st.effects.push(effect);
    }
  }

  tick () {
    if (st.frame > this.last_aa + this.unit.stats.asp*st.fps && this.can_hit(this._target)) {
      this.aa(this._target);
      this.last_aa = st.frame;
    }
  }
}
